"""
Modules dealing with reading data from observations
"""

from .spectrum import BaseSpectrum
from .observed import ObservedSpectrum
from .array import ArraySpectrum
from .taurex import TaurexSpectrum
